-- -*-coding: utf-8;-*-

--[[
-- ************************************************************************
-- BuddyName.lua
-- ************************************************************************
-- This addon is released under GPL v3. All additionally used libraries are
-- released under the license that accompanies them.
--]]

BuddyName = LibStub("AceAddon-3.0"):NewAddon("BuddyName",
                                             "AceConsole-3.0",
                                             "AceHook-3.0",
                                             "AceTimer-3.0")

local L = LibStub("AceLocale-3.0"):GetLocale("BuddyName")

local chatChannels = {
	 ["WHISPER"]      = true,
	 ["PARTY"]        = true,
	 ["GUILD"]        = true,
	 ["YELL"]         = true,
	 ["PARTY"]        = true,
	 ["RAID"]         = true,
	 ["BATTLEGROUND"] = true,
	 ["CHANNEL"]      = false,
}

local battleNetEvents = {
	 ["BN_TOON_NAME_UPDATED"]      = false,
	 ["CHAT_MSG_CHANNEL_LIST"]     = false, -- debug purposes
	 ["BN_CONNECTED"]              = false,
	 ["BN_DISCONNECTED"]           = false,
	 ["BN_FRIEND_TOON_ONLINE"]     = false,
	 ["BN_FRIEND_TOON_OFFLINE"]    = false,
	 ["BN_NEW_PRESENCE"]           = false,
	 ["BN_SELF_ONLINE"]            = false,
	 ["BN_SELF_OFFLINE"]           = false,
}

local slashCommands = {
	 ["bn"]        = true,
	 ["buddyname"] = true,
	 ["BuddyName"] = true,
}

--[[ Local variables ]]
local BN_DB_Version = 2; -- Database format
local BN_CommVersion = 1; -- Message format communication protocol
local BN_Active = false; -- Is this addon active? Modified by OnEnable and OnDisable
local BN_EventFrame = nil; -- Used to register events and process them. Created in OnInitialize
local currentRealm = nil;
local BNetTimer = nil;

--[[ -- -------------------------------- -- ]]
--[[ -- Basic functions, no dependenices -- ]]
--[[ -- -------------------------------- -- ]]

--[[ Creates a new numeric timestamp for the current time and date
--   Returns:
--     1. Numeric: timestamp in format YYYYMMDDHHmm
--]]
function BuddyName:GetTimestamp()
	 local _,month,day,year = CalendarGetDate();
	 local hour,minute = GetGameTime();

	 local timestamp = minute + 100 * (hour + 100 * (day + 100 * (month + 100 * year)))

	 -- self:Print("Debug Timestamp: "..timestamp);

	 return timestamp
end

--[[ Concatenates an arbitrary number of arguments to one string.
--   Only used for debugging purposes.
--   Returns:
--     1. Concatenated version of all supplied arguments
--]]
local function allToString(...)
	 return strjoin(" ", tostringall(...));
end

--[[ -- -------------------------- -- ]]
--[[ -- Database-related functions -- ]]
--[[ -- -------------------------- -- ]]

--[[ Transforms old DB entries to newer versions
--   Returns: nil
--]]
function BuddyName:UpdateDB()
	 if (self.db.realm.version==1) then
			self:Print(string.format(L["UPDATING_DB"], 2));

			-- perform updates to v2
			for p,v in pairs(self.db.realm.char) do
				 if (self.db.realm.char[p]) then
						self.db.realm.char[p] = {
							 buddy = v,
							 guild = "", -- TODO
							 timestamp = BuddyName:GetTimestamp(),
						}
				 end
			end

			self.db.realm.version=2
			self:Print(L["UPDATE_OK"])
	 end

	 --[[
	 if (self.db.realm.version==2) then
			self:Print(string.format(L["UPDATING_DB"], 3));
			-- perform update to v3 here
			self.db.realm.version=3
			self:Print(L["UPDATE_OK"])
	 end
	 ]]
end

--[[ Checks whether the given charname has a buddy name in our DB
--   Returns:
--     1. Boolean: true if author was found in DB, false else
--     2. String: Buddy name of author, or nil if it doesn't exist or equals empty string
--]]
function BuddyName:IsInDB(author)
	 if (author~=nil and author~="") then
			if (self.db.realm.char and -- simplify? test!
					self.db.realm.char[author] and
					self.db.realm.char[author].buddy and
					self.db.realm.char[author].buddy~="")
			then
				 return true,self.db.realm.char[author].buddy;
			end
	 end

	 return false, nil
end

--[[ Adds a new entry to the DB or updates an existing one
--   Returns:
--     1. Boolean: true if successul, false else
--]]
function BuddyName:addToDB(charname, buddyname)
	 print(charname);
	 print(buddyname);
	 self.db.realm.char[charname]= {
			["buddy"] = buddyname,
			["guild"] = "", -- TODO
			["timestamp"] = BuddyName:GetTimestamp()
	 }
	 return true
end

--[[ Deletes an entry from the DB if it exists. Depending on the deleteCompletely flag,
--   an existing entry may either be deleted completely (true) or just set to "" (false).
--   The latter leaves the actual entry intact, resulting in its spreading over the guild
--   synchronization mechanism. Therefore, it is possible to delete this entry from other
--   guild members DBs as well.
--   Returns:
--     1. Boolean: true if deletion was successful, false else
--     2. String: buddy name if deletion was successful, nil else (e.g. char was not in DB)
--]]
function BuddyName:deleteFromDB(charname, deleteCompletely)
	 -- is this char in the DB?
	 if (not BuddyName.db.realm.char[charname]) then
			return false,nil
	 end

	 -- perform deletion and set return value
	 local buddyname = BuddyName.db.realm.char[charname].buddy
	 if (deleteCompletely==true) then
			BuddyName.db.realm.char[charname]=nil
	 else
			BuddyName.db.realm.char[charname].buddy=""
			BuddyName.db.realm.char[charname].timestamp=BuddyName:GetTimestamp();
	 end

	 return true,buddyname
end

function BuddyName:FixDB(version)
	 if (version==2) then
			-- remove all buddy names with "|K...", as they result from broken RealID stuff
			for p,v in pairs(self.db.realm.char) do
				 local pstart = string.find(self.db.realm.char[p].buddy, "^|K");
				 if (pstart) then
						self.db.realm.char[p] = nil;
						self:Print(string.format("Removed broken entry for %s",p));
				 end
			end
	 end
end

--[[ -- -------------- -- ]]
--[[ -- Message output -- ]]
--[[ -- -------------- -- ]]

--[[ Hooked-in version of the AddMessage function.
--   Will check if message contains matching character link,
--   and insert buddy name if match was found.
--   Returns: nil
--]]
function BuddyName:AddMessage(frame, msg, ...)
	 local pstart,pend,sender,type = string.find(msg, "|Hplayer:(%a+):%d+:(%u+)[^|]*|h%[[^%]]+%]"); -- TODO: try to match against |h..|h instead of [..]
	 if (pstart) then -- something matching was found, else nil
			local replace,buddy = BuddyName:IsInDB(sender) -- char in DB?
			if (replace) then
				 if (chatChannels[type]==true or -- msg occured in one of the monitored channels
						string.find(msg,L["ANNOUNCE_ONLINE"]) or -- player has come online
						string.find(msg,L["ANNOUNCE_ACHIEVEMENT"]) ) -- player earned an achievement
				 then
						msg=string.gsub(msg, "(%[[^%]]*)("..sender..")([^%[]*%])", '%1%2:'..buddy..'%3',1); -- only replace first occurrence
				 end
			end
	 else
			pstart,pend,sender=string.find(msg, L["ANNOUNCE_OFFLINE"]);
			if (pstart) then
				 local replace,buddy = BuddyName:IsInDB(sender) -- char in DB?
				 if (replace) then
						msg=string.gsub(msg, L["ANNOUNCE_OFFLINE"], "%1"..buddy, 1);
				 end
			end
	 end
	 
	 -- call original AddMessage() function
	 self.hooks[frame]["AddMessage"](frame, msg, ...);
end

--[[ Adds the current chars of all Battle.net friends that are currently
--   online on the same server, unless there is already an entry for
--   this char.
--   Returns: nil
--]]
function BuddyName:AddBattleNetFriends()
	 local numAll, numOnline = BNGetNumFriends();

	 for i=1, numOnline do -- check all online Battle.net friends
			local _, givenname = BNGetFriendInfo(i); -- does not return the correct givenname, only a battle.net link!
			for j=1, BNGetNumFriendToons(i) do
				 local _, charname, client, realm = BNGetFriendToonInfo(i,j);
				 if (client == "WoW" and realm==currentRealm) then
						local exists,_ = BuddyName.IsInDB(charname); -- most likely does not work correctly here!
						if (exists==false) then
							 --BuddyName.addToDB(charname, givenname);
							 print(string.format(L["BUDDY_ADD_BN_OK"], charname, givenname));
						end
				 end
			end
	 end
end

--[[ Event Handler that is currently only called by Battle.net events.
--   Returns: nil
--]]
local function BattleNetEventHandler(self, event, ...)
	 if (BN_Active == false) then
			return;
	 end

	 if (battleNetEvents[event]==true) then
			-- if BN_FRIEND_TOON_ONLINE or similar, parse event parameters instead of adding all friends
			BuddyName.AddBattleNetFriends();
			--print("Registered event of type "..event); -- debug
	 end
end

--[[ -- ----------------- -- ]]
--[[ -- AddOn interaction -- ]]
--[[ -- ----------------- -- ]]

--[[ Process slash commands (unexpected, eh?)
--   Returns: nil
--]]
function BuddyName:SlashProcessor(input)
	 -- 'input' contains whatever follows the slash command

	 -- split input string, ignore additional whitespace
	 local tbl = {}
	 for v in string.gmatch(input, "[^ ]+") do
			tinsert(tbl, v)
	 end

	 -- if no additional parameters => print out help text
	 if (#tbl == 0) then
			--self:Print(L["PLACEHOLDER"]);
			self:Print(L["HELP"])
			return;
	 end

	 local command = tremove(tbl,1)

	 -- Activate additional functionality
	 if (command=="on") then
			self:Enable()
			self:Print(L["ACTIVATED"]);
			return
	 end

	 -- Disable additional functionality
	 if (command=="off") then
			self:Disable()
			self:Print(L["DEACTIVATED"]);
			return
	 end

	 if (not BN_Active) then
			self:Print(L["NOT_ACTIVE"])
			return
	 end

	 -- add another person to the list
	 if (command=="add") then
			if (#tbl~=2) then
				 self:Print("Add: "..L["WRONG_NUMBER_OF_ARGUMENTS"])
			end
			-- TODO: Check same things as in chatfilter: Is char really player, etc
			BuddyName:addToDB(tbl[1], tbl[2])
			self:Print(string.format(L["BUDDY_ADD_OK"], tbl[1], tbl[2]));
			return
	 end

	 -- add another person to the list
	 if (command=="del" or command=="rm") then
			if (#tbl~=1) then
				 self:Print("Del: "..L["WRONG_NUMBER_OF_ARGUMENTS"])
			end
			local successful,buddyname = BuddyName:deleteFromDB(tbl[1], false)
			if (not successful) then
				 self:Print(string.format(L["BUDDY_DEL_ERROR"], tbl[1]))
				 return
			end
			self:Print(string.format(L["BUDDY_DEL_OK"], tbl[1], buddyname));
			return
	 end

	 -- list chars from database
	 if (command=="list") then
			self:Print(string.format(L["BUDDY_LIST_HEAD"],currentRealm));

			for p,v in pairs(BuddyName.db.realm.char) do
				 local char = p;
				 local buddy = BuddyName.db.realm.char[p].buddy -- DB v2
				 if (buddy and buddy~="") then
						self:Print(string.format(L["BUDDY_LIST_ENTRY"], char, buddy));
				 end
			end

			-- TODO:
			-- Distinguish between all, guild, non-guild
			-- => Additional command arguments
			return
	 end

	 self:Print(string.format(L["UNKNOWN_COMMAND"], command));

-- debug
--	 for v = 1,(#tbl) do
--			self:Print(ChatFrame1, tbl[v]);
--	 end
end

--[[ -- --------------------------------------------- -- ]]
--[[ -- Basic functions automatically called by Ace3  -- ]]
--[[ -- --------------------------------------------- -- ]]

--[[ Executed only once on startup
--   Returns: nil
--]]
function BuddyName:OnInitialize()
	 --self:Print("OnInitialize called"); -- debug

	 currentRealm = GetRealmName();

	 -- Register Database
	 self.db = LibStub("AceDB-3.0"):New("BuddyName_DB")

	 -- Register Slash commands
	 for p,v in pairs(slashCommands) do
			if (v == true) then
				 self:RegisterChatCommand(p, "SlashProcessor")
			end
	 end

	 -- Create frame for events to hook up to and register battle net events
	 BN_EventFrame = CreateFrame("Frame", "BuddyNameFrame");
	 BN_EventFrame:SetScript("OnEvent", BattleNetEventHandler);
	 for p,v in pairs(battleNetEvents) do
			if (v == true) then
				 BN_EventFrame:RegisterEvent(p);
			end
	 end
end


--[[ Executed each time the addon is switched on
--   Returns: nil
--]]
function BuddyName:OnEnable()
	 --self:Print("OnEnable called"); -- debug

	 if (BN_Active == true) then
			return -- we're already active
	 else
			BN_Active = true;
	 end

	 -- check if DB already exists, or if it is old and has to be converted
	 -- not done earlier because has to be called after PLAYER_ENTERING_WORLD (because of UpdateDB())
	 if (self.db.realm.version==nil) then -- if no version number, then either fresh db or very old. In each case, reset realm DB.
			self.db.realm.version=BN_DB_Version -- set current version number
			self.db.realm.char = nil; -- set empty char table
			self.db.realm.char = {};
	 elseif (self.db.realm.version < BN_DB_Version) then -- update old database
			BuddyName:UpdateDB();
	 elseif (self.db.realm.version > BN_DB_Version) then -- database is newer than BuddyName
			self:Print(L["UPDATE_BUDDYNAME"]);
			BN_Active = false;
			return;
	 end

	 -- Apply fixes for broken DB entries
	 if (self.db.realm.version==2) then
			BuddyName:FixDB(2);
	 end

	 -- Hook into AddMessage
	 for i = 1, NUM_CHAT_WINDOWS do
			if (i~=2) then -- exclude combat log, better check against type than number
				 self:RawHook(_G["ChatFrame"..i], "AddMessage", "AddMessage", true);
			end
	 end

	 -- Add one-shot timer to check for battle.net friends in a few seconds
	 --BNetTimer = self:ScheduleTimer("AddBattleNetFriends", 5);

end


--[[ Executed each time the addon is switched off
--   Returns: nil
--]]
function BuddyName:OnDisable()
	 --self:Print("OnDisable called");

	 if (BN_Active == false) then
			return -- we're already inactive
	 else
			BN_Active = false;
	 end

	 -- Unhook from AddMessage
	 self:Unhook("AddMessage");

	 -- Cancel battle.net timer if it has not already fired
	 --self:CancelTimer(BNetTimer, true); -- true: don't raise exception if timer is invalid
end
